@@ -16,25 +16,46 @@
  * Changed to use _int10tostr.
  */
 
+#define _GNU_SOURCE
 #define __FORCE_GLIBC
 #include <features.h>
-#define _STDIO_UTILITY			/* For _int10tostr. */
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
 #include <netinet/in.h>
+#include <bits/uClibc_uintmaxtostr.h>
 
 int inet_aton(const char *cp, struct in_addr *addrptr);
 
 #ifdef L_inet_aton
+/*
+ * More undocumented inet_aton features.
+ * (read: uclibc doesnt support but glibc does)
+ * http://www.mkssoftware.com/docs/man3/inet_aton.3.asp
+ *
+ * *cp can take the form of:
+ * a.b.c.d    - {a,b,c,d} -> 1 byte
+ * a.b.c      - {a,b}     -> 1 byte    {c} -> 2 bytes
+ * a.b        - {a}       -> 1 byte    {b} -> 3 bytes
+ * a          -                        {a} -> 4 bytes
+ *
+ * Each part may be decimal, octal, or hexadecimal as in ISO C.
+ * 0x or 0X    -> hexadecimal
+ * leading 0   -> octal
+ * all else    -> decimal
+ */
 int inet_aton(cp, addrptr)
 const char *cp;
 struct in_addr *addrptr;
 {
-	unsigned long addr;
+	in_addr_t addr;
 	int value;
 	int part;
 
+	if (cp == NULL) {	/* check for NULL arg */
+	    return 0;
+	}
+
 	addr = 0;
 	for (part = 1; part <= 4; part++) {
 
@@ -70,7 +91,7 @@ struct in_addr *addrptr;
 	 *  of the input string, but does not store the result.
 	 */
 	if (addrptr) {
-	    addrptr->s_addr = htonl(addr);
+		addrptr->s_addr = htonl(addr);
 	}
 
 	return 1;
@@ -78,12 +99,12 @@ struct in_addr *addrptr;
 #endif
 
 #ifdef L_inet_addr
-unsigned long inet_addr(const char *cp)
+in_addr_t inet_addr(const char *cp)
 {
 	struct in_addr a;
 
 	if (!inet_aton(cp, &a))
-		return -1;
+		return INADDR_NONE;
 	else
 		return a.s_addr;
 }
@@ -95,10 +116,10 @@ unsigned long inet_addr(const char *cp)
 
 char *inet_ntoa_r(struct in_addr in, char buf[INET_NTOA_MAX_LEN])
 {
-	unsigned long addr = ntohl(in.s_addr);
+	in_addr_t addr = ntohl(in.s_addr);
 	int i;
 	char *p, *q;
-   
+
 	q = 0;
 	p = buf + INET_NTOA_MAX_LEN - 1; /* cannot use sizeof(buf) here */
 	for (i=0 ; i < 4 ; i++ ) {
@@ -125,20 +146,20 @@ char *inet_ntoa(struct in_addr in)
  * Formulate an Internet address from network + host.  Used in
  * building addresses stored in the ifnet structure.
  */
-struct in_addr inet_makeaddr(unsigned long net, unsigned long host)
+struct in_addr inet_makeaddr(in_addr_t net, in_addr_t host)
 {
-        unsigned long addr;
+	in_addr_t addr;
 
-        if (net < 128)
-                addr = (net << IN_CLASSA_NSHIFT) | (host & IN_CLASSA_HOST);
-        else if (net < 65536)
-                addr = (net << IN_CLASSB_NSHIFT) | (host & IN_CLASSB_HOST);
-        else if (net < 16777216L)
-                addr = (net << IN_CLASSC_NSHIFT) | (host & IN_CLASSC_HOST);
-        else
-                addr = net | host;
-        addr = htonl(addr);
-        return (*(struct in_addr *)&addr);
+	if (net < 128)
+		addr = (net << IN_CLASSA_NSHIFT) | (host & IN_CLASSA_HOST);
+	else if (net < 65536)
+		addr = (net << IN_CLASSB_NSHIFT) | (host & IN_CLASSB_HOST);
+	else if (net < 16777216UL)
+		addr = (net << IN_CLASSC_NSHIFT) | (host & IN_CLASSC_HOST);
+	else
+		addr = net | host;
+	addr = htonl(addr);
+	return (*(struct in_addr *)&addr);
 }
 
 #endif
@@ -149,9 +170,9 @@ struct in_addr inet_makeaddr(unsigned long net, unsigned long host)
  * internet address; handles class a/b/c network
  * number formats.
  */
-unsigned long inet_lnaof(struct in_addr in)
+in_addr_t inet_lnaof(struct in_addr in)
 {
-	unsigned long i = ntohl(in.s_addr);
+	in_addr_t i = ntohl(in.s_addr);
 
 	if (IN_CLASSA(i))
 		return ((i)&IN_CLASSA_HOST);
@@ -168,17 +189,17 @@ unsigned long inet_lnaof(struct in_addr in)
  * Return the network number from an internet
  * address; handles class a/b/c network #'s.
  */
-u_int32_t
+in_addr_t
 inet_netof(struct in_addr in)
 {
-        u_int32_t i = ntohl(in.s_addr);
+	in_addr_t i = ntohl(in.s_addr);
 
-        if (IN_CLASSA(i))
-                return (((i)&IN_CLASSA_NET) >> IN_CLASSA_NSHIFT);
-        else if (IN_CLASSB(i))
-                return (((i)&IN_CLASSB_NET) >> IN_CLASSB_NSHIFT);
-        else
-                return (((i)&IN_CLASSC_NET) >> IN_CLASSC_NSHIFT);
+	if (IN_CLASSA(i))
+		return (((i)&IN_CLASSA_NET) >> IN_CLASSA_NSHIFT);
+	else if (IN_CLASSB(i))
+		return (((i)&IN_CLASSB_NET) >> IN_CLASSB_NSHIFT);
+	else
+	return (((i)&IN_CLASSC_NET) >> IN_CLASSC_NSHIFT);
 }
 
 #endif
