@@ -304,6 +304,10 @@ long __XL(strtol)(const char * __restrict str, char ** __restrict endptr,
 
 __XL_ALIAS(strtol)
 
+#if defined(__sparc__) // to be able to use libstdc++ ontop 
+weak_alias (__XL(strtol),__strtol_internal)
+#endif
+     
 #endif
 /**********************************************************************/
 #if defined(L_strtoll) || defined(L_strtoll_l)
@@ -327,6 +331,10 @@ long long __XL(strtoll)(const char * __restrict str,
 
 __XL_ALIAS(strtoll)
 
+#if defined(__sparc__) // to be able to use libstdc++ ontop 
+weak_alias (__XL(strtoll),__strtoll_internal)
+#endif
+     
 #endif /* defined(ULLONG_MAX) && (LLONG_MAX > LONG_MAX) */
 
 #endif
@@ -350,6 +358,10 @@ unsigned long __XL(strtoul)(const char * __restrict str,
 
 __XL_ALIAS(strtoul)
 
+#if defined(__sparc__) // to be able to use libstdc++ ontop 
+weak_alias (__XL(strtoul),__strtoul_internal)
+#endif
+     
 #endif
 /**********************************************************************/
 #if defined(L_strtoull) || defined(L_strtoull_l)
@@ -372,6 +384,10 @@ unsigned long long __XL(strtoull)(const char * __restrict str,
 
 __XL_ALIAS(strtoull)
 
+#if defined(__sparc__) // to be able to use libstdc++ ontop 
+weak_alias (__XL(strtoull),__strtoull_internal)
+#endif
+     
 #endif /* defined(ULLONG_MAX) && (LLONG_MAX > LONG_MAX) */
 
 #endif
@@ -863,7 +879,7 @@ int mblen(register const char *s, size_t n)
 	size_t r;
 
 	if (!s) {
-		state.mask = 0;
+		state.__mask = 0;
 #ifdef __CTYPE_HAS_UTF_8_LOCALES
 		return ENCODING == __ctype_encoding_utf8;
 #else
@@ -873,7 +889,7 @@ int mblen(register const char *s, size_t n)
 
 	if ((r = mbrlen(s, n, &state)) == (size_t) -2) {
 		/* TODO: Should we set an error state? */
-		state.wc = 0xffffU;		/* Make sure we're in an error state. */
+		state.__wc = 0xffffU;	/* Make sure we're in an error state. */
 		return (size_t) -1;		/* TODO: Change error code above? */
 	}
 	return r;
@@ -889,7 +905,7 @@ int mbtowc(wchar_t *__restrict pwc, register const char *__restrict s, size_t n)
 	size_t r;
 
 	if (!s) {
-		state.mask = 0;
+		state.__mask = 0;
 #ifdef __CTYPE_HAS_UTF_8_LOCALES
 		return ENCODING == __ctype_encoding_utf8;
 #else
@@ -899,7 +915,7 @@ int mbtowc(wchar_t *__restrict pwc, register const char *__restrict s, size_t n)
 
 	if ((r = mbrtowc(pwc, s, n, &state)) == (size_t) -2) {
 		/* TODO: Should we set an error state? */
-		state.wc = 0xffffU;		/* Make sure we're in an error state. */
+		state.__wc = 0xffffU;	/* Make sure we're in an error state. */
 		return (size_t) -1;		/* TODO: Change error code above? */
 	}
 	return r;
@@ -932,7 +948,7 @@ size_t mbstowcs(wchar_t * __restrict pwcs, const char * __restrict s, size_t n)
 	mbstate_t state;
 	const char *e = s;			/* Needed because of restrict. */
 
-	state.mask = 0;				/* Always start in initial shift state. */
+	state.__mask = 0;			/* Always start in initial shift state. */
 	return mbsrtowcs(pwcs, &e, n, &state);
 }
 
