@@ -230,6 +230,10 @@ static int current_rtmax = -1;
 int __pthread_sig_restart = SIGUSR1;
 int __pthread_sig_cancel = SIGUSR2;
 int __pthread_sig_debug;
+
+void (*__pthread_restart)(pthread_descr) = __pthread_restart_old;
+void (*__pthread_suspend)(pthread_descr) = __pthread_suspend_old;
+int (*__pthread_timedsuspend)(pthread_descr, const struct timespec *) = __pthread_timedsuspend_old;
 #else
 
 #if __SIGRTMAX - __SIGRTMIN >= 3
@@ -376,6 +380,7 @@ static void pthread_initialize(void)
   sa.sa_flags = 0;
   __libc_sigaction(__pthread_sig_restart, &sa, NULL);
   sa.sa_handler = pthread_handle_sigcancel;
+  sigaddset(&sa.sa_mask, __pthread_sig_restart);
   // sa.sa_flags = 0;
   __libc_sigaction(__pthread_sig_cancel, &sa, NULL);
   if (__pthread_sig_debug > 0) {
@@ -388,6 +393,10 @@ static void pthread_initialize(void)
   sigemptyset(&mask);
   sigaddset(&mask, __pthread_sig_restart);
   sigprocmask(SIG_BLOCK, &mask, NULL);
+  /* And unblock __pthread_sig_cancel if it has been blocked. */
+  sigdelset(&mask, __pthread_sig_restart);
+  sigaddset(&mask, __pthread_sig_cancel);
+  sigprocmask(SIG_UNBLOCK, &mask, NULL);
   /* Register an exit function to kill all other threads. */
   /* Do it early so that user-registered atexit functions are called
      before pthread_onexit_process. */
@@ -508,7 +517,7 @@ int __pthread_initialize_manager(void)
       __pthread_wait_for_restart_signal(thread_self());
     }
   /* Synchronize debugging of the thread manager */
-PDEBUG("send REQ_DEBUG to manager thread\n");
+  PDEBUG("send REQ_DEBUG to manager thread\n");
   request.req_kind = REQ_DEBUG;
   TEMP_FAILURE_RETRY(__libc_write(__pthread_manager_request,
 	      (char *) &request, sizeof(request)));
@@ -535,9 +544,9 @@ int pthread_create(pthread_t *thread, const pthread_attr_t *attr,
   PDEBUG("write REQ_CREATE to manager thread\n");
   TEMP_FAILURE_RETRY(__libc_write(__pthread_manager_request,
 	      (char *) &request, sizeof(request)));
-PDEBUG("before suspend(self)\n");
+  PDEBUG("before suspend(self)\n");
   suspend(self);
-PDEBUG("after suspend(self)\n");
+  PDEBUG("after suspend(self)\n");
   if (THREAD_GETMEM(self, p_retcode) == 0)
     *thread = (pthread_t) THREAD_GETMEM(self, p_retval);
   return THREAD_GETMEM(self, p_retcode);
@@ -572,7 +581,7 @@ pthread_descr __pthread_find_self()
 
 #ifdef DEBUG_PT
   if (h->h_descr == NULL) {
-      printf("*** "__FUNCTION__" ERROR descriptor is NULL!!!!! ***\n\n");
+      printf("*** %s ERROR descriptor is NULL!!!!! ***\n\n", __FUNCTION__);
       _exit(1);
   }
 #endif
@@ -889,7 +898,7 @@ __pthread_timedsuspend_old(pthread_descr self, const struct timespec *abstime)
 	struct timespec reltime;
 
 	/* Compute a time offset relative to now.  */
-	__gettimeofday (&now, NULL);
+	gettimeofday (&now, NULL);
 	reltime.tv_nsec = abstime->tv_nsec - now.tv_usec * 1000;
 	reltime.tv_sec = abstime->tv_sec - now.tv_sec;
 	if (reltime.tv_nsec < 0) {
